page 60,132
;------------------------------------------------------------------------
;
;			D  S  K  C  M  P
;
;------------------------------------------------------------------------
;
;	ACTION:	This program compares two floppy diskettes, similar to
;		IBM utility DISKCOMP.
;
;	REVISION HISTORY:
;
;	Version 1.0 pre-production release
;
;	10/12/84  NF  partially coded
;
;     	10/27/84/3:45PM
;       Modified and updated by Vuong Tran
;
;		There are several restrictions in this initial version:
;
;***** verify this first claim!
;		1.  Like the IBM version, it does not allow comparison
;		    between different media types.  This is no problem
;		    with Apollo, but with Venus this means we cannot
;		    compare two logically identical but physically
;		    dissimilar media (like 3.5" and 5.25" diskettes in
;		    DSDD format).
;
;		2.  No Quad density support (80 tpi).
;
;		3.  No clever programming to reduce head settle time,
;		    which IBM uses.
;
;	12/07/84
;		1.  Quad density support (80 tracks)
;
;		2.  Substitute DOS call 1CH (get FAT ID - does not work
; 	    properly with mixing 40/80 tracks mode) by DOS call
;		    25H (absolute disk read).
;
;	12/17/84
;		Restore the use of DOS call 1CH (get FAT ID). The bug in
;		<IO.SYS> has been fixed for mixing 40/80tracks mode.
;		Fix a bug in <read_cylinder> routine
;
;------------------------------------------------------------------------
;
;----- global symbols ----------------------------------------------------

;----- equates -----------------------------------------------------------

; version number

version		equ	10h

; interrupt numbers

warm_boot	equ	20h
DOS_fct		equ	21h

; DOS functions

DOS_print_string	equ	09h
DOS_clear_kbd_and_wait	equ	0CH
DOS_switchar		equ	37H
DOS_fat_id		equ	1CH
get_default_drive	equ	19h

; miscellaneous
true			equ	0
false			equ	0FFH
cr			equ	13
lf			equ	10
p_s_p_top_memory	equ	02H
u_p_a_length		equ	80H
u_p_a			equ	81H
invalid_drive		equ     0FFH
single_side_switch	equ     '1'
single_side_format	equ     1H
eight_sectors_switch	equ     '8'
eight_sectors_format	equ     2H
eighty_tracks_format	equ	04h
sector_size_bytes	equ	512
drive1_num		equ	5CH	;1st formatted parm area
drive2_num		equ	6CH	;2nd formatted parm area

;-------------------------------------------------------------------------

code		segment
		assume  cs:code, ds:code, es:code, ss:code
		org 	100h
diskcomp:	jmp 	start

;----- data area ---------------------------------------------------------



				db 	version
drives		label	word
drive1				db	(?)
drive2				db	(?)
default_drive			db	?

first_time			db	?
switches			db	?
compare_format			db	?
next_diskette			db	(?)
current_diskette		db	?
cylinder			db	?	;valid cylinders in range 0..39
diskette1_cylinder		db	?
diskette2_cylinder		db	?
max_cylinder			db	?
last_cylinder			db	?
compare_cylinder		db	?
retry_read			db	?
cmp_error_flag			db	?	;if any error set flag to 0ffh

buffer1_base_abs		dw	?
buffer1_top_abs			dw	?
buffer2_base_abs		dw	?
buffer2_top_abs			dw	?
free_space_base_abs		dw	?
free_space_top_abs		dw	?
cylinder_base_abs		dw	?
cylinder_top_abs		dw	?
cylinder_size_paras		dw	?
cylinder_size_bytes		dw	?

	;------ Messages
insert_diskette1	db	'Please insert first diskette into drive '
drive_x			db	'x',cr,lf,cr,lf,'$'
insert_diskette2	db	'Please insert second diskette into drive '
drive_y			db	'y',cr,lf,cr,lf,'$'
waiting_msg		db	'Strike any key when ready...',cr,lf,cr,lf,'$'
repeat_msg		db	'Compare more diskettes (Y/N)?','$'
bad_format_msg		db	'Incompatible diskettes...',cr,lf,cr,lf,'$'
bad_switch_msg		db	'Bad parameters...',cr,lf,cr,lf,'$'
invalid_drive_msg	db	'Invalid drive specification...',cr,lf,cr,lf,'$'

not_ready_msg		db	'Drive '
not_ready_drive		db	'  not ready',cr,lf,cr,lf,'$'

format_type_msg		db	'Comparing '
num_of_sectors		db	'9 sectors per track, '
num_of_sides		db	'2 side(s), '
format_density		db	'40 tracks',cr,lf,cr,lf,'$'

unrecover_msg		db	'Unrecoverable read error on drive '
unrecover_drive		db	' ',cr,lf,cr,lf,'$'

location_error		db	'Side '
error_side		db	0,' track '
error_track		db 	0,0,cr,lf,cr,lf,'$'

cmp_error_msg		db	'Compare error on',cr,lf,'$'
skip_line		db	cr,lf,cr,lf,'$'
compare_ok_msg		db	'Diskettes compare OK',cr,lf,cr,lf,'$'

;fat_error		db	'Cannot read FAT ID',cr,lf,cr,lf,'$'

	;FAT buffer
;fat_buffer		db	512 dup (?)	;buffer for FAT sector



;-------------------------------------------------------------------------
;
;	IMPLEMENTATION NOTES:
;
;	Algorithm below assumes normal case is a single drive copy,
;	which requires user to swap diskettes out of same drive.
;	Double drive copy is handled as special case.
;	Algorithm minimizes need for user to swap diskettes.
;

	;* Beginning of Main Program

start:		mov	sp, offset tos	;set up local stack
		call	initial		; handle entry conditions from OS
		call	check_version
		call	check_equipment
		call	get_drives
		call	get_switches
		call	get_memory

again:		call	insert_first_diskette
		call	check_diskette1
		call	init_loop_variables

loop:		call	fill_buffer
		call	insert_next_diskette
		call	check_diskette2
		call	read_and_compare_buffer
		call	update_loop_variables
		jnz	loop
                cmp	[cmp_error_flag], 0ffh
		jz	atr_next
		mov	dx, offset compare_ok_msg
		call	print_string
atr_next:	call	ask_to_repeat
		jmp	again

	;* End of Main Program


;-------------------------------------------------------------------------

print_string		proc	near

;	ENTRY:	[dx]	ptr to string to be printed, delimited by "$".
;
;	ACTION:	Print string.
;
;	EXIT:	All registers unchanged.

		mov	ah, DOS_print_string
		int	DOS_fct
		ret

print_string	endp

;-------------------------------------------------------------------------

initial		proc	near

;	ENTRY:	[ax]	drive numbers of formal parameters
;			as defined by MS-DOS on entry
;
;	ACTION:	Handle entry conditions from MS-DOS.
;		Initialize drive number variables.
;
;	EXIT:	All registers destroyed.
 
		mov 	[drives], ax	; save drive numbers from entry
		ret
initial		endp


;-------------------------------------------------------------------------

check_version		proc	near

;	ENTRY:	Do not care.
;
;	ACTION:	Check DOS version number is correct for this program.
;
;	EXIT:	All registers destroyed.

;***** disabled since we will always be greater than 2.0 DOS level.
;*
;*		mov 	ah, DOS_version
;*		int 	DOS_fct
;*		cmp 	al, 02		; verify DOS major rev level >= 2
;*		jae 	ok1
;*		mov 	dx, version_err_msg
;*		call 	abort
;*************************************
ok1:		ret

check_version		endp

;-------------------------------------------------------------------------

check_equipment		proc	near

;	ENTRY:	Do not care.
;
;	ACTION:	Perform equipment check for disk drives being present.
;		Create bit map of allowable drives.
;
;	EXIT:	All registers destroyed.

;***** unimplemented. Apollo needs not be implemented.

		ret

check_equipment		endp

;-------------------------------------------------------------------------

get_drives		proc	near

;	ENTRY:	Do not care.
;
;	ACTION:	Get drive numbers to compare.
;		Force drive1 and drive2 to default drive if need be.
;		Force single drive compare if need be.

;***** couldn't default drive be ram disk or hard disk???
;
;	EXIT:	All registers destroyed.

		cmp	[drive1], invalid_drive
		jz	error1
		cmp	[drive2], invalid_drive
		jz	error1

;********** concerns:
;***** 1.  relax above criterion to force to default drive??
;***** check IBM!
;***** 2.  Check whether this supports device reassignment, like VENUS
;***** CMOS data area, DEV_ASSIGN.
;***** 3.  Will VENUS have to check that external floppies are physically
;***** identical? what about drives?
;***** 4.  Write drive letters into messages.


	;Determine user selected drives. VQT
		mov	ah, get_default_drive	;get the default drive
		int	DOS_fct			;default now in AL
		mov	[default_drive], al	;save for later
		mov	si, drive1_num 		;points to formatted parm area @5CH
		lodsb				;fetch 1st drive designation
		or	al, al			;see if drive having been specified
		jnz	drvspec			;if specified proceed
		mov	al, [default_drive]
		inc	al
drvspec:	dec	al			;drive designator now correct
		mov	byte ptr ds:[drive1_num], al	;and updated (0=A, 1=B, etc)
		mov	drive1, al
		add	al,'A'
		mov	[drive_x], al		;update message

		mov	si, drive2_num 		;points to formatted parm area @6CH
		lodsb				;fetch 2nd drive designation
		or	al, al			;see if drive having been specified
		jnz	drv2spec		;if specified proceed
		mov	al, [default_drive]
		inc	al
drv2spec:	dec	al			;drive designator now correct
		mov	byte ptr ds:[drive2_num], al	;and updated (0=A, 1=B, etc)
		mov	drive2, al
		add	al, 'A'
		mov	[drive_y], al           ;update message
		ret

error1:		mov	dx, offset invalid_drive_msg
		call	abort

get_drives		endp

;-------------------------------------------------------------------------

get_switches		proc	near

;	ENTRY:	Do not care.
;
;	ACTION:	Get switches from unformatted parameter area.  Set
;		corresponding bits in <switches> to indicate operational modes.
;
;	EXIT:	All registers destroyed.


		mov	[switches], 0
		mov	ah, DOS_switchar	; read switch character
		mov	al, 0 		;
		int	DOS_fct			; using undocumented fct
						; which returns it DL.

		mov	bx, u_p_a_length	; set string length in bytes
		mov	cl, byte ptr [bx]	;
		sub	ch, ch
		mov	al, dl			; target = switchar
		mov	di, u_p_a		; set string begin

repeat:		repnz	scasb			; while di <> eos
						; scan for target in string
		jcxz	end_of_string		; jump when eos, even if
						; matches switchar, since
						; switch name missing.
		cmp	byte ptr [di], single_side_switch
		jne	next1
		or	[switches], single_side_format
		jmp	repeat

next1:		cmp	byte ptr [di], eight_sectors_switch
		jne	bad_switch_name
		or	[switches], eight_sectors_format	
		jmp	repeat
bad_switch_name:
		mov	dx, offset bad_switch_msg
		call	abort

end_of_string:	ret


get_switches		endp

;-------------------------------------------------------------------------

get_memory		proc	near

;	ENTRY:	Do not care.
;
;	ACTION:	Find top and bottom of free space.
;		Check against minimum needed by algorithm.
;		Abort with error if not enough space.
;
;	EXIT:	All registers destroyed.

		mov	bx, offset free_space_base_rel
		mov	cl, 4
		shr	bx, cl		; convert to paragraphs
		inc	bx		; compensate in case truncated.
					; (overflow impossible)
		mov	ax, cs
		add	ax, bx		; convert to absolute address
		mov	[free_space_base_abs], ax
		mov	bx, p_s_p_top_memory
		mov	ax, [bx]
		mov	[free_space_top_abs], ax

; Minimum space needed is 2 cylinders.  Worst case means two sides of
; two tracks of 9 sectors.  Hence 18K of space needed.

;***** unimplemented check and may actually not be needed, even in
;*****	OAK with 64K!

		ret

get_memory		endp

;-------------------------------------------------------------------------

insert_first_diskette		proc	near

;	ENTRY:	Do not care.
;
;	ACTION:	Ask user to insert first diskette.
;
;		Special case: if double drive copy, ask him to insert
;		second diskette also.
;
;	EXIT:	All registers destroyed.


		mov	dx, offset insert_diskette1
		call	print_string
		mov	al, [drive2]
		cmp	[drive1], al
		je	skip1				;VQT
		mov	dx, offset insert_diskette2	; double drive operation
		call	print_string

skip1:		call	wait_for_godot
		ret

insert_first_diskette		endp


;-------------------------------------------------------------------------

wait_for_godot		proc	near

;	ENTRY:	Do not care.
;
;	ACTION:	Print "waiting..." message and
;		wait for Mr. Godot to press any key.
;
;	EXIT:	All registers destroyed.
;		(AL) = user input char


		mov	dx, offset waiting_msg
		call	print_string
		mov	ah, DOS_clear_kbd_and_wait	; wait for keystroke
		mov	al, 08h				;console input without echo. VQT
		int	DOS_fct                         ;input char returned in (AL)
		ret

wait_for_godot	endp
;-------------------------------------------------------------------------

check_diskette1		proc	near

;	ENTRY:	Do not care.
;
;	ACTION:	Check first diskette format and set comparison format
;
;		Special case:  Check second diskette format only if double
;		drive operation.  Compare diskette format.
;
;	EXIT:	All registers destroyed.

		mov	al, [drive1]
		call	check_diskette_format
		call	set_compare_format	; set format for comparison
		mov	al, [drive1]
		cmp	al, [drive2]
		je	single3
		mov	al, [drive2]		; double drive case
		call	check_diskette_format
		call	compare_diskette_formats; verify diskette2 agrees
						; with comparison format.

single3:	ret

check_diskette1		endp

;-------------------------------------------------------------------------

init_loop_variables		proc	near

;	ENTRY:	Do not care.
;
;	ACTION:	Initialize variables used by routines in loop.
;
;	EXIT:	All registers destroyed.

		mov	[first_time], true		; first time thru loop flag
		mov	[current_diskette], 1
		mov	[next_diskette], 2		; next diskette to insert
		mov	ax, sector_size_bytes
		shl	ax, 1				; multiply by 8
		shl	ax, 1
		shl	ax, 1
		test	[compare_format], eight_sectors_format
		jnz	eight
		add	ax, sector_size_bytes		; nine sectors

eight:		test	[compare_format], single_side_format
		jnz	single
		add	ax, ax				;double sided

single:		mov	[cylinder_size_bytes], ax
		shr	ax, 1                   	;divide by 16
		shr	ax, 1				;
		shr	ax, 1				;
		shr	ax, 1				;
		mov	[cylinder_size_paras], ax

; initialize buffers:
;	size of buffer1 is 1 cylinder, buffer2 is above buffer1 and takes
;	up the rest of free space. Warning! Buffer2 size is not multiple
;	of cylinder size.
                mov	ax, [free_space_base_abs]
		mov	[buffer1_base_abs], ax
		add	ax, [cylinder_size_paras]
		mov	[buffer1_top_abs], ax
		mov	[buffer2_base_abs], ax
		mov	ax, free_space_top_abs
		mov	[buffer2_top_abs], ax

; initialize loop index and limit:

		mov	[diskette1_cylinder], 0
		mov	[diskette2_cylinder], 0

		mov	[max_cylinder], 40	;cylinder after last cylinder
		test	[compare_format], eighty_tracks_format
		jz	mode40			;go if 40 tracks mode
		mov	[max_cylinder], 80	;set 80 tracks mode
mode40:
		mov	[retry_read], 0
		mov	[cmp_error_flag], 0

		ret

init_loop_variables		endp

;-------------------------------------------------------------------------

fill_buffer		proc	near

;	ENTRY: Do not care
;
;	ACTION:	Fill cylinder buffer for given drive.
;
;	EXIT:  All destroyed

	;get next cylinder number
		cmp	[current_diskette], 2
		je	fill1
		mov	al, [diskette1_cylinder]
		mov	[cylinder], al
		mov	bl, [drive1]		;setup drive for read_cylinder
		jmp	fill2

fill1:		mov	al, [diskette2_cylinder]
		mov	[cylinder], al
		mov	bl, [drive2]            ;setup drive for read cylinder

	;initialize cylinder base and top
fill2:		mov	ax, [buffer2_base_abs]
		mov	[cylinder_base_abs], ax
                mov	ax, [cylinder_base_abs]
		mov	[cylinder_top_abs], ax
fill_loop:	mov	ax, [cylinder_size_paras]
		add	[cylinder_top_abs], ax
                mov	ax, [buffer2_top_abs]
		cmp	[cylinder_top_abs], ax
		jae	fill_done
		mov	al, [max_cylinder]
		cmp	[cylinder], al
		je	fill_done			;[cylinder] is next cylinder to compare

;/* check if 64k boundary will be crossed by next cylinder in buffer2
; if so, DMA from diskette input cannot be used.

		mov	ax, [cylinder_base_abs]
		xor	ax, [cylinder_top_abs]
		test	ax, 1000h		; 64k bytes = 4k paras
		jnz	cross

; will not cross, read directly into buffer2
;*/
		mov	ax, [cylinder_base_abs]
		call	read_cylinder
		jmp	fill3

;/* will cross, so first read into buffer1 and copy into buffer 2
; note: buffer1 is empty at this time.

cross:		mov	ax, [buffer1_base_abs]
		call	read_cylinder

	;copy from buffer1 to cylinder_base in buffer2.
		push	ds
		push	es
		xor	di, di
		xor	si, si
		mov	cx, [cylinder_size_bytes]
		mov 	ax, [cylinder_base_abs]
		mov	es, ax
		mov	ax, [buffer1_base_abs]
		mov	ds, ax
		cld
	rep	movsb
		pop	es
		pop	ds

fill3:
		inc	[cylinder]
		mov	ax, [cylinder_top_abs]
		mov	[cylinder_base_abs], ax
		jmp	fill_loop
fill_done:	mov	al ,[cylinder]
		cmp	[current_diskette],2
		je	fill4
		mov	[diskette1_cylinder], al
		jmp	fill_quit
fill4:		mov	[diskette2_cylinder], al
fill_quit:	ret

fill_buffer		endp

;-------------------------------------------------------------------------

insert_next_diskette		proc	near

;	ENTRY:	Do not care.
;
;	ACTION:	Ask user to insert next diskette and wait.
;		Special case: double drive
;
;	EXIT:	All registers destroyed.

		mov	al, [drive1]		;2 drives case
		cmp	al, [drive2]            ;
		jnz	ind_quit                ;

		cmp	[next_diskette], 2
		je	two_disk
		mov	dx, offset insert_diskette1
		call	print_string
		jmp	wait

two_disk:	mov	dx, offset insert_diskette2
		call	print_string

wait:		call	wait_for_godot
ind_quit:	ret

insert_next_diskette		endp

;-------------------------------------------------------------------------

check_diskette2		proc	near

;	ENTRY:
;
;	ACTION:	Check second diskette format, if first time thru loop.
;		Compare diskette formats.
;
;		Special case:  if double drive operation, then do nothing,
;		since already checked.
;
;	EXIT:

		cmp	[first_time], false
		je	not_first_time
		push	ax
		mov	al, [drive1]
		cmp	al, [drive2]
		pop	ax
		jne	not_first_time          ;go if double drive operation
		mov	al, [drive2]		; single drive compare
		call	check_diskette_format
		call	compare_diskette_formats

not_first_time:	ret

check_diskette2		endp

;-------------------------------------------------------------------------

read_and_compare_buffer		proc	near

;	ENTRY:
;
;	ACTION:	Read and compare cylinders from [next_diskette] drive with
;		already full buffer.
;
;	EXIT:
		mov	ax, [buffer2_base_abs]
                mov	[cylinder_base_abs], ax
		cmp	[next_diskette], 2
		je	read2
		mov	al, [diskette2_cylinder]
		mov	[compare_cylinder], al
		mov	bl, al
		mov	al, [diskette1_cylinder]
		mov	[cylinder], al			;VQT
		mov	[diskette1_cylinder], bl
		mov	bl, [drive1]
		jmp	read_loop

read2:         mov	al, [diskette1_cylinder]
		mov	[compare_cylinder], al
		mov	bl, al
		mov	al, [diskette2_cylinder]
		mov	[cylinder], al
		mov	[diskette2_cylinder], bl
		mov	bl, [drive2]

read_loop:      mov	al, [compare_cylinder]
		cmp	[cylinder], al
		jge	read_done

		mov	ax, [buffer1_base_abs]
		call	read_cylinder
                mov 	ax, [cylinder_base_abs]
		call	cmp_cylinder
		inc	[cylinder]
		mov	ax, [cylinder_size_paras]
		add	[cylinder_base_abs], ax
		jmp	read_loop

read_done:	ret

read_and_compare_buffer		endp

;-------------------------------------------------------------------------

update_loop_variables		proc	near

;	ENTRY:	Do not care.
;
;	ACTION:	Update loop variables and return flag when loop done.
;
;	EXIT:	(ZF)	= 0 if not done
;			= 1 if done
;		All other registers and flags destroyed.

		mov	[first_time], false	; already through loop once
		mov	al, [diskette1_cylinder]
		cmp	[max_cylinder], al	;diskette1 cylinder
 						     ;equals diskette2 cylinder now
		je	update_done
		xor	[current_diskette], 3	; toggles current_diskette
		xor	[next_diskette], 3	; toggles next diskette
		

update_done:	ret

update_loop_variables		endp

;-------------------------------------------------------------------------

ask_to_repeat		proc	near

;	ENTRY:	Do not care.
;
;	ACTION:	Ask user if he wants to repeat operation and return
;		if he does.  Otherwise warm boot system.	
;
;	EXIT:	All registers destroyed.


		mov	dx, offset repeat_msg
		call	print_string
		mov	ah, DOS_clear_kbd_and_wait
		mov	al, 01h		;keyboard input with echo
		int	DOS_fct		;input char in (AL)
		or	al, 20h		;convert to lower case
		cmp	al, 'y'
		jz	atr_quit        ;go if want to repeat
 					;the whole process again
		int	warm_boot       ;else, goback to DOS
atr_quit:       mov	dx, offset skip_line
		call	print_string
		ret

ask_to_repeat	endp

;-------------------------------------------------------------------------

abort		proc	near

;	ENTRY:	[dx]	points to error message, delimited by "$".
;
;	ACTION:	Print out fatal error message and warm boot.
;
;	EXIT:	Do not care since control never returns to caller!

		call	print_string
		int	warm_boot
		ret

abort		endp

;-------------------------------------------------------------------------

check_diskette_format		proc	near

;	ENTRY:	[al]	drive number (0=A).
;
;	ACTION:	Check actual format of diskette in given drive.
;		Is it nine sectored or is second side formatted,
;		or is it 80/40 tracks formatted?
;
;	EXIT:	[al]	flag indicating actual format
;			(bit fields correspond to switches)
;		restore DS
;		All other registers destroyed.


	;* Substitute DOS call 1CH by 25H
;		MOV	CX, 4 		;SET # OF RETRIES
;RETRY:		PUSH	CX
;		PUSH	AX
;		MOV	BX, OFFSET FAT_BUFFER	;DS:BX --> FAT_BUFFER
;		MOV	CX, 01			;1 SECTOR READ
;		MOV	DX, 01			;BEGINNING RELATIVE SECTOR
;		INT	25H			;READ 1ST FAT SECTOR
;		POPF				;NECESSARY
;		POP	AX
;		POP	CX
;		JNC	SUCCESSFUL
;		LOOP	RETRY
;		MOV	DX, OFFSET FAT_ERROR
;		CALL	ABORT			;BACK TO DOS

;SUCCESSFUL:	MOV	AL, BYTE PTR FAT_BUFFER	;GET FAT ID

		mov	dl, al
		inc	dl		;drive designator now correct
		mov	ah, DOS_fat_id	;allocation table information
		int	DOS_fct		;
		mov	al, ds:[bx]	;get FAT ID (ds here has new value)
		push	cs		;restore ds
		pop	ds		;


		cmp	al, 0ffh
		jz	double_sided_8
		cmp	al, 0feh
		jz	single_sided_8
		cmp	al, 0fdh
		jz	double_sided_9
		cmp	al, 0fch
		jz	single_sided_9

		CMP	AL, 0FBH
		JZ	ds8_80tracks
		CMP	AL, 0FAH
		JZ	ss8_80tracks
		CMP	AL, 0F9H
		JZ	ds9_80tracks

		mov	dx, offset bad_format_msg
		call	print_string
		int	warm_boot
double_sided_8:	mov	al, 02h
		ret
single_sided_8:	mov	al, 03h
		ret
double_sided_9:	mov	al, 00h
		ret
single_sided_9:	mov	al, 01h
		ret
ds8_80tracks:	mov 	al, 06h
		ret
ss8_80tracks:	mov	al, 07h
		ret
ds9_80tracks:	mov	al, 04h
		ret

check_diskette_format		endp

;-------------------------------------------------------------------------

set_compare_format		proc	near

;	ENTRY:	[AL] format of diskette 1
;
;	ACTION:	Set format in which compare operation will be done.
;		Print out message about this.
;
;	EXIT:	All registers destroyed.

;***** check IBM logic, verify that diskettes are treated symmetrically.

		or	al, [switches]	;switches override actual format. VQT
		mov	[compare_format], al

;***** implemented print out of "comparing...", see IBM @ 431 and 49f. VQT
		mov	[num_of_sides], '2'	;double sided
		shr	al, 1
		jnc	sectors_per_track
		mov	[num_of_sides], '1'	;single sided
sectors_per_track:
		mov	[num_of_sectors], '9'	;9 sectors/track
		shr	al, 1
		jnc	density
		mov	[num_of_sectors], '8'   ;8 sectors per track
density:	mov	[format_density], '4'   ;40 tracks mode
		shr	al, 1
		jnc	format_type
		mov	[format_density], '8'   ;80 tracks mode
format_type:	mov	dx, offset format_type_msg
		call	print_string
		ret

set_compare_format		endp

;-------------------------------------------------------------------------

compare_diskette_formats	proc	near

;	ENTRY:	[al] format of second diskette
;
;	ACTION:	Compare second diskette format with comparison format,
;		and switches.  Switches override native format of both
;		diskettes.  Abort with error message if they are unequal.
;
;	EXIT:	All registers destroyed.

		or	al, [switches]		; switches override. VQT
		cmp	al, [compare_format]
		je	ok3
		mov	dx, offset bad_format_msg
		call	abort

ok3:		ret

compare_diskette_formats	endp

;-------------------------------------------------------------------------

read_cylinder		proc	near

;	ENTRY:	[ax]	absolute address of where to put cylinder, in paras.
;		[bl]	drive number to read from
;
;	ACTION:	Read cylinder [cylinder] from diskette drive into
;		given address.  Do max of 4 retrys, if failure report
;		to user and return.
;
;	EXIT:	[ax], [bl] preserved, [bh]=head#
;**** see IBM @44d, @4a4

		mov	cl, compare_format
		cmp	cl, 0 	;single-density
		jz	ds9
		cmp	cl, 1           ;single-density
		jz	ss9
		cmp	cl, 2           ;single-density
		jz	ds8
		cmp	cl, 3           ;single-density
		jz	ss8
		cmp	cl, 4           ;quad-density
		jz	ds9
		cmp	cl, 6           ;quad-density
		jz	ds8

	;case single sided, 8 sectors/track (cl=3 or 7)
	;* read 1 track (track # = cylinder #)from side 0, 8 sectors
ss8:		sub	bh, bh			;head 0
		push	ax
		push	bx
		mov	ch, 8			;8 sectors
		call	do_read_cylinder
		call	read_error      	;any read error?
		pop	bx
		pop	ax
		jc	ss8             	;retry if read error (will loop
						;     forever if timeout error)
		jmp     rc_quit                 ;go if no read error or
						;read error with retries exhausted

	;case single sided, 9 sectors/track (cl=1)
	;* read 1 track (track # = cylinder #) from side 0, 9 sectors
ss9:		sub	bh, bh			;head 0
		push	ax
		push	bx
		mov	ch, 9			;9 sectors
		call	do_read_cylinder
		call	read_error
		pop	bx
		pop	ax
		jc	ss9
		jmp     rc_quit

	;case double sided, 8 sectors/track  (cl=2 or 6)
		;* read 1 track (track # = cylinder #)from side 0, 8 sectors
ds8:		sub	bh, bh			;head 0
	        push	ax
		push	bx
		mov	ch, 8			;8 sectors
		call	do_read_cylinder
		call	read_error
		pop	bx
		pop	ax
		jc	ds8

		;* update buffer pointer
		add	ax, 100h

		;* read same track, but from side 1
ds8_side1:	mov	bh, 01h         	;head 1
		push	ax
		push	bx
		mov	ch, 8			;8 sectors
		call	do_read_cylinder
		call	read_error
		pop	bx
		pop	ax
		jc	ds8_side1
		jmp     rc_quit

	;case double sided, 9 sectors/track (cl=0 or 4)
		;* read 1 track (track # = cylinder #) from side 0, 9 sectors
ds9:		sub	bh, bh			;head 0
                push	ax
		push	bx
		mov	ch, 9			;9 sectors
		call	do_read_cylinder
		call	read_error
		pop	bx
		pop	ax
		jc	ds9

		;* update buffer pointer
		add	ax, 120h

		;* read same track, but from side 1
ds9_side1:	mov	bh, 01h         	;head 1
          	push	ax
		push	bx
		mov	ch, 9			;9 sectors
		call	do_read_cylinder
		call	read_error
		pop	bx
		pop	ax
		jc	ds9_side1
rc_quit:	ret
read_cylinder	endp


;-------------------------------------------------------------------------

do_read_cylinder	proc	near

;	ENTRY:
;		bh = head #
;		bl = drive #
;		ch = number of sectors to be read
;		ax = absolute address of buffer
;		<cylinder>
;
;	ACTION: read 1 track
;
;	EXIT:   ds,es preserved
;		dh = head #
;		dl = drive #
;		ch = track #
;		cl = start sector #
;		ah = status of operation
;		CY = 0 successful read (ah=0 on return)
;		CY = 1 failed operation (ah has error reason)

                push	es
		mov	dl, bl 		;get drive number

		mov	dh, bh		;head #
		mov	es, ax		;es:bx ---> buffer
		xor	bx, bx		;
		mov	al, ch		;total number of sectors in al
		mov	ch, [cylinder]	;track #
		mov	cl, 01h		;sector 1
		mov	ah, 02h		;read disk

		PUSH	DX		;REMEMBER DRIVE, HEAD
		TEST	[COMPARE_FORMAT], EIGHTY_TRACKS_FORMAT
		JZ	MODE_40TRACKS
		OR	DL, 40H		;SET UP FOR 80 TRACKS MODE READ
MODE_40TRACKS:
		INT	13H
		POP	DX		;RESTORE DRIVE, HEAD

		pop	es
		ret
do_read_cylinder	endp
read_error	proc	near
;	ENTRY:
;		dh = head #
;		dl = drive #
;		ch = track #
;		cl = start sector #
;		ah = diskette status
;		(CF) = 1 if read error
;		(CF) = 0 if no read error
;
;	ACTION:
;		Display appropriate error messages
;
;	EXIT:
;		(CF) = 1 if read error
;		(CF) = 0 if no read error, or
;			    read error with retries exhausted

		jnc	re_quit
		push	ax		;save diskette status
		push	cx
		push	dx
		sub	ax, ax		;reset diskette
		int	13h             ;
		pop	dx
		pop	cx
		pop	ax		;restore diskette status
		cmp	ah, 80h		;time out error?
		jnz	no_time_out
		mov	[not_ready_drive], dl		;update error message
		add	[not_ready_drive], 'A'          ;
		mov	dx, offset not_ready_msg        ;
		call	print_string                    ;
read_again:	stc
		ret
re_quit:	mov	[retry_read], 0
		ret

no_time_out:	inc	retry_read
		cmp	retry_read, 04h
		jnz	read_again			;retry reading disk
		mov	word ptr [error_track], 3030h	;convert track # to ASCII
convert2:	cmp	ch, 0ah                         ;
		jb	convert1                        ;
		inc	byte ptr [error_track]          ;
		sub	ch, 0ah                         ;
		jmp	convert2                        ;
convert1:	add	byte ptr [error_track]+1, ch    ;

		add	dh, 30h			;convert side # to ASCII
 		mov	[error_side], dh
		mov	[unrecover_drive], dl
		add	[unrecover_drive], 'A'

		mov	dx, offset unrecover_msg
		call	print_string

		mov	dx, offset location_error
		call	print_string

		clc
		jmp	re_quit
read_error	endp

;-------------------------------------------------------------------------
cmp_cylinder	proc	near

;	ENTRY:
;		 [AX] = <cylinder_base_abs>
;		 [buffer1_base_abs]
;		 [cylinder_size_bytes]
;		 [cylinder]
;
;	ACTION:  Comparing buffer pointed by [ax] with buffer
;		 pointed by <buffer1_base_abs>. Comparing size
;		 is 1 cylinder.
;
;	EXIT:   report error if any, or else quit

		mov	cx, [cylinder_size_bytes]
		mov	si, [buffer1_base_abs]
		test	[compare_format],single_side_format
		jz	two_sides
		call	compare_track		;do side 0
		jz	cmp_cyl_quit
		mov	[cmp_error_flag], 0ffh
		mov	[error_side], '0'
		call	report_track_error
		jmp	cmp_cyl_quit

two_sides:      shr	cx, 1
 		call	compare_track           ;do side 0
		jz	do_side1
		mov	[cmp_error_flag], 0ffh
		mov	[error_side], '0'
		call	report_track_error

		;* update pointer
do_side1:  	mov	cx, [cylinder_size_bytes]
		shr	cx, 1               ;get half bytes left
		mov	di, cx 	    	    ;convert to paragraphs
		shr	di, 1               ;
		shr	di, 1               ;
		shr	di, 1               ;
		shr	di, 1               ;
		mov	ax, [cylinder_base_abs]
		add	ax, di
		mov	si, [buffer1_base_abs]  ;ds:si points to buffer base
		add	si, di
		call	compare_track           ;do side 1
		jz	cmp_cyl_quit
		mov	[cmp_error_flag], 0ffh
		mov	[error_side], '1'
		call	report_track_error
cmp_cyl_quit:	ret
cmp_cylinder	endp

;----------------------------------------------
compare_track	proc
;entry:	cx =size of buffer in bytes for comparison
;	ax = absolute address of cylinder base
;	si = absolute address of buffer base
;exit: 	(ZF) =1 IF no error
;	(ZF) =0 IF ERROR

 		push	ds
		push	es
		mov	es, ax    	;es:di points to cylinder base
		mov	ax, si		;ds:si points to buffer base
		mov	ds, ax
try_next_side:	xor	si, si
		xor	di, di
		cld
	repz	cmpsb
		pop	es
		pop	ds
		ret
compare_track	endp


report_track_error	proc

		mov	dh, [cylinder]	            ;convert track # to ASCII
		mov	word ptr [error_track],3030h;
convert4:	cmp	dh, 0ah                     ;
		jb	convert3                    ;
		inc	byte ptr [error_track]      ;
		sub	dh, 0ah                     ;
		jmp	convert4                    ;
convert3:	add	byte ptr [error_track]+1, dh;

		mov	dx, offset cmp_error_msg
		call	print_string

		mov	dx, offset location_error
		call	print_string
		ret
report_track_error	endp






	; STACK AREA
		DB	 80H DUP (?)
	TOS	LABEL	BYTE


free_space_base_rel:			; beginning of free space
code		ends
		end 	diskcomp
