PAGE	60,132
;---------------------------------------------------------------:
;								:
; FILENAME: DSKHDAXT.ASM ---- TURNQUIST SPECIAL			:
;								:
;	CUSTOM FORMAT, NO DIAGNOSTICS TEST			:
;								:
;	9/13/84							:
;								:
;DSKHD: HARD DISK BIOS ROUTINE FOR QX88 & APOLLO-X. SUPPORTS    :
;       THE COMREX HARD DISK. WHAT FOLLOWS BELOW IS THE         :
;       SPECIFICATION FOR THIS ROUTINE.                         :
;                                                               :
;INPUT  (AH = HEX VALUE)                                        :
;                                                               :
;    AH=00 RESET DISK (DL = 80H,81H)/DISKETTE                   :
;       01 READ THE STATUS OF THE LAST DISK OPERATION           :
;          NOTE: DL < 80H - DISKETTE                            :
;                DL > 80H - DISK                                :
;       02 READ THE DESIRED SECTORS INTO MEMORY                 :
;       03 WRITE THE DESIRED SECTORS FROM MEMORY                :
;       04 VERIFY THE DESIRED SECTORS                           :
;       05 FORMAT THE DESIRED TRACK                             :
;       06 FORMAT THE DESIRED TRACK AND SET BAD SECTOR FLAGS    :
;       07 FORMAT THE DRIVE STARTING AT THE DESIRED TRACK       :
;       08 RETURN THE CURRENT DRIVE PARAMETERS                  :
;       09 INITIALIZE DRIVE PAIR CHARACTERISTICS (NOT        )  :
;           INTERRUPT 41 POINTS TO DATA BLOCK    (IMPLEMENTED)  :
;       0A READ LONG                                            :
;       0B WRITE LONG                                           :
;       0C SEEK                                                 :
;       0D ALTERNATE DISK RESET (SEE DL) (NOT IMPLEMENTED)      :
;       0E READ SECTOR BUFFER   (NOT IMPLEMENTED)               :
;       0F WRITE SECTOR BUFFER  (NOT IMPLEMENTED)               :
;       10 TEST DRIVE READY                                     :
;       11 RECALIBRATE                                          :
;       12 CONTROLLER RAM DIAGNOSTIC                            :
;       13 DRIVE DIAGNOSTIC                                     :
;       14 CONTROLLER INTERNAL DIAGNOSTIC                       :
;                                                               :
;       REGISTERS USED FOR FIXED DISK OPERATIONS                :
;                                                               :
;       DL - DRIVE NUMBER       (80H-87H FOR DISK, VAL CHECKED) :
;       DH - HEAD NUMBER        (0-7 ALLOWED, NOT VAL CHECKED)  :
;       CH - CYLINDER NUMBER    (0-1023, NOT VAL CHECKED)       :
;       CL - SECTOR NUMBER      (1-17, NOT VAL CHECKED)         :
;                                                               :
;           NOTE: HIGH 2 BITS OF CYLINDER NUMBER ARE PLACED     :
;                 IN THE HIGH 2 BITS OF THE CL REGISTER         :
;                 ( 10 BITS TOTAL)                              :
;                                                               :
;       AL - NUMBER OF SECTORS (MAXIMUM POSSIBLE RANGE 1-80H,   :
;                               FOR READ/WRITE LONG 1-79H)      :	
;            (INTERLEAVE VALUE FOR FORMAT)                      :
;               (NOT USED IN QX-88/APOLLO)                      :
;       ES:BX - ADDRESS OF BUFFER FOR READS AND WRITES.         :
;               NOT USED FOR VERIFY.                            :
;                                                               :
; OUTPUT                                                        :
;       AH = STATUS OF CURRENT OPERATION                        :
;            STATUS BITS ARE DEFINED IN EQUATES BELOW           :
;       CY = 0 SUCCESSFUL OPERATION (AH=0 ON RETURN)            :
;       CY = 1 FAILED OPERATION (AH HAS ERROR REASON)           :
;                                                               :
;       NOTE:  ERROR 11H INDICATES THAT THE DATA READ HAD A     :
;       RECOVERABLE ERROR WHICH WAS CORRECTED BY THE ECC        :
;       ALGORITHM. THE DATA IS PROBABLY GOOD, HOWEVER THE BIOS  :
;       ROUTINE INDICATES AN ERROR TO ALLOW THE CONTROLLING     :
;       PROGRAM A CHANCE TO DECIDE FOR ITSELF. THE ERROR MAY    :
;       NOT RECUR IF THE DATA IS REWRITTEN. (AL) CONTAINS THE   :
;       BURST LENGTH (NOT IMPLEMENTED IN QX88/APOLLO)           :
;                                                               :
;       IF DRIVE PARAMETERS WERE REQUESTED,                     :
;       RETURNS FIXED VALUES USED FOR COMREX HARD DISK, DOES    :
;       NOT USE TABLES AS IN THE IBM.                           :
;                                                               :
;       DL = NUMBER OF CONSECUTIVE ACKNOWLEDGING DRIVES         :
;            ATTACHED (0-2) (0 FOR CONTROLLER CARD ONLY)        :
;       DH = MAXIMUM USEABLE VALUE FOR HEAD NUMBER              :
;       CH = MAXIMUM USEABLE VALUE FOR CYLINDER NUMBER          :
;       CL = MAXIMUM USEABLE VALUE FOR SECTOR NUMBER            :
;            AND CYLINDER NUMBER HIGH BITS                      :
;                                                               :
;       REGISTERS WILL BE PRESERVED EXCEPT WHEN THEY ARE USED   :
;       TO RETURN INFORMATION                                   :
;                                                               :
;       NOTE: IF AN ERROR IS REPORTED BY THE DISK CODE, THE     :
;       APPROPRIATE ACTION IS TO RESET THE DISK, THEN RETRY     :
;       THE OPERATION.                                          :
;                                                               :
;---------------------------------------------------------------:
;ADDITIONAL NOTES REGARDING QX-88/APOLLO IMPLEMENTATION:        :
;                                                               :
;(1) Since the hard disk controller board has its own buffer,   :
;    the bios routine runs in non-dma mode, allowing the occur- :
;    rence of interrupts during the transfer without loss of    :
;    data.                                                      :
;                                                               :
;(2) Getparm returns a fixed set of parameters to be used with  :
;    the COMREX hard disk and associated controller. Since this :
;    is the only device to be used with the QX-88/Apollo, the   :
;    use of tables was skipped.                                 :
;                                                               :
;(3) The interleave for the format is fixed at 8:1 as in the    :
;        CP/M BIOS for the hard disk.                           :
;                                                               :
;(4) Only additional BIOS variable used is DSTAT (hard disk     :
;    status byte). This is being placed where IBM has motor     :
;    status byte because QX-88/Apollo do not use this byte.     :
;    Other command byte area used by IBM for setting up their   :
;    command stream to their controller is not needed/used in   :
;    this implementation due to the use of r/w registers pro-   :
;    vided by the hdc used with the COMREX.                     :
;                                                               :
;(5) No change is needed in this routine from QX-88 to Apollo   :
;    implementations. No dependent code used.                   :
;                                                               :
;---------------------------------------------------------------:
;           * * * IMPLEMENTATION HISTORY * * *                  :
;                                                               :
;     DATE                                                      :
;    6/28/84: Start conversion from 'C'                         :
;                                                               :
;    7/04/84: Finish conversion from 'C'                        :
;                                                               :
;    7/13/84: Successful test (1) booting system from floppy.   :
;                             (2) format hard using IBM format. :
;                             (3) file transfer from/to hard.   :
;                                                               :
;    7/14/84: Code walk-through, finished! (for now?)           :
;                                                               :
;    1/10/85: MADE LOADABLE DEVICE DRIVER. DS SCREW-UP IN       :
;             ROM CODE (SHIT!). ADDED FIX (THX EDDIE).          :
;                                                               :
;                                                               :
;---------------------------------------------------------------:
PAGE
;	ERROR EQUATES
;
SENSE_FAIL	EQU	0FFH	;SENSE OPERATION FAILED
UNDEF_ERR	EQU	0BBH	;UNDEFINED ERROR OCCURRED
HTIME_OUT 	EQU	80H	;ATTACHMENT FAILED TO RESPOND
HBAD_SEEK 	EQU	40H	;SEEK OPERATION FAILED
BAD_CNTLR  	EQU	20H	;CONTROLLER HAS FAILED
DATA_CORRECTED	EQU	11H	;ECC CORRECTED DATA ERROR
BAD_ECC  	EQU	10H	;BAD ECC ON DISK READ
BAD_TRACK	EQU	0BH	;BAD TRACK FLAG DETECTED
DMA_BOUNDARY 	EQU	09H	;ATTEMPT TO DMA ACROSS 64K BOUNDARY
INIT_FAIL	EQU	07H	;DRIVE PARAMETER ACTIVITY FAILED
BAD_RESET	EQU	05H	;RESET FAILED
RECORD_NOT_FND  EQU	04H	;REQUESTED SECTOR NOT FOUND
HBAD_ADDR_MARK 	EQU	02H	;ADDRESS MARK NOT FOUND
HBAD_CMD  	EQU	01H	;BAD COMMAND PASSED TO DISK I/O
;
; Port equates for F-01 HARD DISK CONTROLLER
;
WDBASE		EQU	0080H		;  Base QX 10 port of F-01 	
HDATA		EQU	WDBASE		;  Data port 	(DATA in 'C' vers)
WPC		EQU	WDBASE+1 	;  Write precomp port 
HARD_ERROR		EQU	WPC	 	;  Error port 
SECNT		EQU	WDBASE+2 	;  Sector count 
SECNO		EQU	WDBASE+3 	;  Sector number
CYLLO		EQU	WDBASE+4 	;  Cylinder low 
CYLHI		EQU	WDBASE+5 	;  Cylinder high
SDH		EQU	WDBASE+6 	;  Size/Drive/Head 
COMND		EQU	WDBASE+7 	;  Command register
STATUS		EQU	COMND	 	;  Status register 
;
; Command equates for F-01 HARD DISK CONTROLLER
;
CTEST		EQU	090H	; Command: Test
CREST		EQU	010H	; Command: Restore               & r3 r2 r1
CSEEK		EQU	070H 	; Command: Seek                  & r3 r2 r1
CREAD		EQU	020H	; Command: Read sector           &  M  L
CREADD		EQU	028H	; Command: Read sector DMA mode  &  M  L
CWRITE  	EQU	030H	; Command: Write sector          &  M  L
CFORM   	EQU	050H	; Command: Format track 		
MULTIPLE 	EQU	004H	;   extra: Multiple sector operation	
LONG		EQU	002H	;   extra: Long operation		

SECSIZE 	EQU	512	; bytes per sector 	
MAX_FILE	EQU	0	; MAXIMUM NUMBER OF DISK DRIVES (0 BASED)



PAGE

HDROUT	PROC	FAR
	CMP	DL,80H		;TEST FOR FIXED DISK DRIVE
	JZ	HARD_DISK	;YES HANDLE HERE - COMREX HD DRIVE=80=0
 
  	INT	40H		;DISKETTE HANDLER
RET_2:
 	RET	2		;RETURN FROM DISKETTE HANDLER
 
HARD_DISK:			;ENTRY POINT FOR HARD DISK IO
 	STI			;ENABLE INTERRUPTS
	OR	AH,AH
	JNZ	ABC3

  	INT	40H		;RESET NEC WHEN AH = 0
	SUB	AH,AH
	CMP	DL,80H 	     
	JNZ	RET_2		;NOT COMREX HARD DISK, SO RETURN
ABC3:
 	CMP	AH,08		;GET PARAMETERS IS A SPECIAL CASE
 	JNZ	ABC2
 	JMP	GET_PARM_N	;GO GET PARMS, RETURN FROM THERE
ABC2:
	PUSH	BX		;SAVE REGISTERS DURING OPERATION
	PUSH	CX
	PUSH	DX
	PUSH	DS
	PUSH	ES
	PUSH	SI
	PUSH	DI
	CALL	DISK_IO_CONT	;MAIN "BRANCH" ROUTINE
	PUSH	AX		;SAVE STATUS RETURN AX REG
	MOV	AL,030H		
	OUT	SDH,AL		;DESELECT DRIVE
;	ASSUME	DS:DATA
 	MOV	AX,DATA		;ESTABLISH DATA SEGMENT
 	MOV	DS,AX
	POP	AX		;RESTORE STATUS RETURN AX REG
	MOV	AH,DS:DSTAT	;RETURN STATUS
	CMP	AH,1		;SET CARRY IF AH IS NON-ZERO (ERROR COND)
	CMC
	POP	DI		;RESTORE REGISTERS (EXCEPT FOR AX)=RETURN
	POP	SI
	POP	ES
	POP	DS
	POP	DX
	POP	CX
	POP	BX
 	RET	2	;PRESERVE FLAGS REGISTER ON RETURN FROM INTERRUPT
HDROUT	ENDP

PAGE
DISK_IO_CONT	PROC	NEAR	;MAIN BRANCHING ROUTINE
 	PUSH	AX
 	MOV	AX,DATA	;ESTABLISH SEGMENT
 	MOV	DS,AX
 	POP	AX
	CMP	AH,01H	;RETURN STATUS COMMAND?
	JNZ	ABC4
	JMP	RETURN_STATUS
ABC4:
	SUB	DL,80H		;CONVERT DRIVE NUMBER TO 0 BASED RANGE
	DEC	CL		;CONVERT SECTORS 0 - 16D
	CMP	DL,MAX_FILE	;LEGAL DRIVE TEST
	JNZ	BAD_COMMAND

	MOV	DS:DSTAT,0		;SET STATUS TO OK
	PUSH	AX
	MOV	AL,AH
	XOR	AH,AH
	SAL	AX,1
	MOV	SI,AX
	CMP	AX,H1L		;CHECK TO SEE IF PAST END OF TABLE
	POP	AX
	JNB	BAD_COMMAND	;COMMAND NOT FOUND, ERROR
	JMP	WORD PTR CS:[SI + OFFSET H1] ;EXECUTE COMMAND
BAD_COMMAND:
	MOV	DS:DSTAT,HBAD_CMD	;ILLEGAL COMMAND, ERROR
	MOV	AL,0
	RET
DISK_IO_CONT	ENDP
PAGE
;---------------------------------------------------------------:
; RESET THE DISK SYSTEM (AH = 000H)                             :
; On IBM, reset to controller is performed. QX-88/Apollo cannot :
; be done. Therefore, perform a recalibrate   as next best type :
; of "reset" to perform.                                        :
;                                                               :
;---------------------------------------------------------------:
DISK_RESET	PROC	NEAR
	CALL	HRECAL		;PERFORM RESTORE
	RET
DISK_RESET	ENDP
;---------------------------------------------------------------:
; DISK STATUS ROUTINE (AH = 001H)                               :
; Same as IBM. return status byte in AL, reset status to 0=ok.  :
;---------------------------------------------------------------:
RETURN_STATUS	PROC	NEAR
	MOV	AL,DS:DSTAT
	MOV	DS:DSTAT,0H
	RET
RETURN_STATUS	ENDP
;---------------------------------------------------------------:
; DISK READ ROUTINE (AH = 002H)                                 :
;---------------------------------------------------------------:
DISK_READ	PROC	NEAR
	MOV	AH,CREAD
	JMP	HDREAD
DISK_READ	ENDP
;---------------------------------------------------------------:
; DISK WRITE ROUTINE (AH = 003H)                                :
;---------------------------------------------------------------:
DISK_WRITE	PROC	NEAR
	MOV	AH,CWRITE
	JMP	HDWRITE
DISK_WRITE	ENDP
;---------------------------------------------------------------:
; DISK VERIFY ROUTINE (AH = 004H)                               :
;---------------------------------------------------------------:
DISK_VERF	PROC	NEAR
	JMP	HDVERF
DISK_VERF	ENDP
;---------------------------------------------------------------:
; FORMATTING (AH = 005H, 006H, 007H)                            :
; AH = Format byte (MSBit represents toggle to format to end of :
;      disk. AH is left shifted in HFORMAT to determine if      :
;      format is to end of track. After shift occurs, the result:
;      value in AH is the format byte: 0=normal format, 1=format:
;      bad track.                                               :
;      This conforms to the IBM standard for interface to their :
;      hard disk. (i.e. even though we have the capability to   :
;      "bad" sector format, we will do "bad" track format as    :
;      does IBM).                                               :
;---------------------------------------------------------------:
FMT_TRK	PROC	NEAR
	XOR	AH,AH 	;NORMAL FORMAT 	(MSBIT=0) (FMT BYTE=0)
	JMP	SHORT FMT_IT
FMT_TRK	ENDP

FMT_BAD	PROC	NEAR
	MOV	AH,040H	;FORMAT ONE BAD SECTOR	(MSBIT=0) (FMT BYTE=80H)
	JMP	SHORT FMT_IT  
FMT_BAD	ENDP

FMT_DRV	PROC	NEAR
	MOV	AH,080H ;FORMAT	(TO END OF DRIVE MSBIT=1) (FMT BYTE=0)
FMT_IT:
	JMP	HFORMAT
FMT_DRV	ENDP
PAGE
;---------------------------------------------------------------:
; GET DRIVE PARAMETERS (AH = 008H)                              :
; Returns fixed parameters for COMREX hard disk, since it is    :
; the only one currently supported.                             :
;                                                               :
;---------------------------------------------------------------:
GET_PARM_N	LABEL	NEAR
GET_PARM	PROC	FAR 
			
	PUSH	DS		;SAVE CURRENT DATA SEGMENT
	PUSH	ES
	PUSH	BX
;	ASSUME	DS:ABS0
	SUB	AX,AX
	MOV	DS,AX
	LES	BX,DS:HF_TBL_VEC
;	ASSUME	DS:DATA
	MOV	AX,DATA
	MOV	DS,AX

	MOV	AX,ES:[BX]	;MAX NUMBER OF CYLINDERS
	SUB	AX,2		; ADJUST FOR 0-N (& RESERVE LAST TRACK)
	MOV	CH,AL		
	AND	AX,0300H	;HIGH 2 BITS OF CYL
	SHR	AX,1
	SHR	AX,1
	OR	AL,010H		;SECTORS
	MOV	CL,AL

	MOV	DH,ES:[BX][2]	;GET HEADS
	DEC	DH		;0-N RANGE
	MOV	DL,1		;CONSECUTIVE DRIVES = 1
EX_GET:
	POP	BX
	POP	ES
	POP	DS		;RESTORE CURRENT DATA SEGMENT
	RET	2		;RETURN FROM INTERRUPT 
GET_PARM  	ENDP
;---------------------------------------------------------------:
;INIT DRV (AH = 009H) (Not used in QX-88/Apollo)                :
;---------------------------------------------------------------:
INIT_DRV	PROC	NEAR
	RET
INIT_DRV	ENDP
;---------------------------------------------------------------:
; READ LONG (AH = 0AH)  Same as READ w/4 ECC bytes per transfer :
;---------------------------------------------------------------:
RD_LONG	PROC	NEAR
	MOV	AH,(CREAD OR 02H)
	JMP	HDREAD
RD_LONG	ENDP
;---------------------------------------------------------------:
; WRITE LONG (AH = 0BH) Same as WRITE w/4 ECC bytes per transfer:
;---------------------------------------------------------------:
WR_LONG	PROC	NEAR
	MOV	AH,(CWRITE OR 02H)
	JMP	HDWRITE
WR_LONG	ENDP	
PAGE
;---------------------------------------------------------------:
; SEEK (AH = 0CH) Sets Step rate and performs Seek              :
;---------------------------------------------------------------:
DISK_SEEK	PROC	NEAR
	CALL	UPTASK
	JMP	ERRCHK
DISK_SEEK	ENDP
;---------------------------------------------------------------:
; READ SECTOR BUFFER (AH = 0EH) Not used in QX-88/Apollo        :
;---------------------------------------------------------------:
RD_BUFF	PROC	NEAR
RD_BUFF	ENDP
;---------------------------------------------------------------:
; WRITE SECTOR BUFFER (AH = 00FH) Not used in QX-88/Apollo      :
;---------------------------------------------------------------:
WR_BUFF	PROC	NEAR
	RET
WR_BUFF	ENDP
;---------------------------------------------------------------:
; TEST DISK READY (AH = 010H)                                   :
;---------------------------------------------------------------:
TST_RDY	PROC	NEAR
	JMP	HREADY
TST_RDY	ENDP
;---------------------------------------------------------------:
; RECALIBRATE (AH = 011H)                                       :
;---------------------------------------------------------------:
HDISK_RECAL	PROC	NEAR
	JMP	HRECAL
HDISK_RECAL	ENDP
;---------------------------------------------------------------:
; All diagnostics use the same routine. Whereas IBM has separate:
; calls for each of their diagnostics in their controller. The  :
; controller with the COMREX has one call to implement them all.:
;---------------------------------------------------------------:
;---------------------------------------------------------------:
; CONTROLLER RAM DIAGNOSTICS (AH = 012H)                        :
;---------------------------------------------------------------:
RAM_DIAG	PROC	NEAR
RAM_DIAG	ENDP
;---------------------------------------------------------------:
; DRIVE DIAGNOSTICS (AH = 013H)                                 :
;---------------------------------------------------------------:
CHK_DRV	PROC	NEAR
CHK_DRV	ENDP
;---------------------------------------------------------------:
; CONTROLLER INTERNAL DIAGNOSTICS (AH = 014H)                   :
;---------------------------------------------------------------:
CNTLR_DIAG	PROC	NEAR
	CLC			;RETURN OK 
	RET			;TURNQUIST 9/13/84 JMP	HDIAG
CNTLR_DIAG	ENDP

PAGE
;---------------------------------------------------------------:
;SUPPORT ROUTINES BELOW                                         :
;---------------------------------------------------------------:
;---------------------------------------------------------------:
;HREADY: See if hard disk is even connected. 6/28/84            :
;        test controller busy, drive ready.                     :
;        if error, change dstat, and set carry flag             :
;   ENTRY: NONE                                                 :
;    EXIT: CY=0 OK, CY=1 ERROR STATUS BYTE SET                  :
;DESTROYS: AL                                                   :
;---------------------------------------------------------------:
HREADY	PROC	NEAR			
	MOV	AL,0A0H		
	OUT	SDH,AL			;SEND OUT TO REGISTER
	PUSH	BX			;SAVE BX REGISTER	
	MOV	BL,AL
	NOT   	AL
	OUT	CYLLO,AL 		;TOGGLE LINE
	IN	AL,SDH			;REGISTER STILL THERE?
	CMP	AL,BL
	JNZ	ERRCOND			;NO, REGISTER NOT THERE, ERROR

	IN	AL,STATUS		;IS CONTROLLER BUSY?
	TEST	AL,040H			;TEST DRIVE READY BIT
	JZ	ERRCOND
;
DRSET:
	TEST	AL,080H			;TEST CONTROLLER BUSY BIT
	JZ	HEXIT   		;IF SET THEN ERROR CONDITION
;
ERRCOND:
	MOV	DS:DSTAT,SENSE_FAIL
	STC
HEXIT:	
	POP    	BX			;RESTORE BX REGISTER
	RET
HREADY	ENDP

;---------------------------------------------------------------:
;CMDREADY: Wait till busy bit is reset. means command finished  :                                                             :
;   ENTRY: NONE                                                 :
;    EXIT: WHEN BUSY BIT IS RESET (WILL EXIT ROUTINE)           :
;DESTROYS: AL                                                   :
;                                                               :
;---------------------------------------------------------------:
CMDREADY	PROC	NEAR	;Test hdc busy bit, (ret if reset)
	IN	AL,STATUS
	TEST	AL,080H		;test BUSY bit
	JNZ	CMDREADY	;if set loop
	RET
CMDREADY	ENDP

PAGE
;---------------------------------------------------------------:
;HRECAL: Recalibrate the hard disk.                             :                                 :
;   ENTRY: NONE                                                 :
;    EXIT: CY=0 OK, CY=1 ERROR                                  :
;DESTROYS: AL                                                   :
;---------------------------------------------------------------:
HRECAL	PROC	NEAR		;RECALIBRATE THE HARD DISK (#0 & #11)
	CALL	HREADY
	JC	NTRDY		;7/12:RFS	
	MOV	AL,CREST	;SEND OUT RESTORE COMMAND
	OUT	COMND,AL
	CALL	CMDREADY	;WAIT TILL HDC IS READY 
	CALL	ERRCHK		;CHECK FOR ERRORS
NTRDY:				;7/12:RFS
	RET
HRECAL	ENDP


PAGE
;---------------------------------------------------------------:
;UPTASK: Set up task files using variables in memory.           :
;       MOVE REGS TO MEMORY VARS, LOAD TASK FILES.              :
;                                                               :
;ENTRY: AH = INTERNAL COMMAND	(REMAINS UNTOUCHED		:
;       AL = SECTOR COUNT                                       :
;	CL = (2MSBITS=CYLHI, 6 OTHER=SECTOR NUMBER START	:
;	CH = CYLINDER LOW					:
;	DL = DRIVE NUMBER					:
;	DH = HEAD NUMBER					:
;                                                               :
;Sets up TASK FILES on hard disk controller:                    :
;       SDH: b7:ECC, b6,b5:Sector size(01=512) b4,b3:drive(00=0):
;            b2,b1,b0:head                                      :
;     CYLLO: Cylinder low value                                 :
;     CYLHI: Cylinder hi  value                                 :
;     SECNO: Starting sector number                             :
;     SECNT: Count of sectors to read/write.                    :
;                                                               :
;Performs "seek", seeking to desired location.                  :
;Sets up the step rate.                                         :
;								:
;EXIT:	AH = INTERNAL COMMAND (REMAINS UNTOUCHED)               :
;	CL = CYLINDER LOW					:
;	CH = CYLINDER HI					:
;	DH = SECTOR NUMBER START				:
;	DL = SECTOR COUNT    					:
;       BX = REMAINS UNTOUCHED			        	:
;                                                               :
;DESTROYED: AL                                                  :
;                                                               :
;Registers are now set up for further processing by read/write  :
; routines.                                                     :
;---------------------------------------------------------------:
UPTASK	PROC	NEAR
	PUSH	AX
	CALL	HREADY		;CHECK TO SEE IF DISK IS READY
	POP	AX
	JC	EXIT1
	AND	DL,003H		;ONLY 2 BINARY DIGIT DRIVE POSSIBILITY
	SHL	DL,1
	SHL	DL,1
	SHL	DL,1		;SHIFT IT LEFT 3 PLACES
	OR	DL,DH		;OR IN HEAD NUMBER
	OR	DL,0A0H		;OR IN SECTOR SIZE=512/ECC SET
	XCHG	DL,AL		;PUT SECNT INTO DL, 
				;PUT SDH INTO AL FOR OUTPUT TO TASK FILE
	OUT	SDH,AL		;SET SDH 
	MOV	AL,CH
	OUT	CYLLO,AL	;SET CYLLO	
	MOV	AL,CL		;SAVE CL TO AL TO SAVE CYLINDER NUMBER BITS
	AND	CL,03FH		;MASK OUT 2 CYL HI BITS
	MOV	DH,CL		;MOVE SECNO TO DH REG
	AND	AL,0C0H		;MASK OUT LOW 6 BITS TO GET CYLLHI IN TOP 2
	MOV	CL,6
	SHR	AL,CL		;MOVE CYLLHI TO LOW 2 BITS OF CL
	MOV	CL,AL
	XCHG	CH,CL		;CX = (CH=CYLHI,CL=CYLLO)
	OUT	CYLHI,AL	;SET CYLHI
	MOV	AL,DH
	OUT	SECNO,AL	;SET SECNO
	MOV	AL,01H		;LOAD SECOUNT W/ 1 SECTOR
	OUT	SECNT,AL
	MOV	AL,CSEEK	;LOAD COMND: TO DO SEEK (& SET STEP RATE)
	OUT	COMND,AL
	CALL	CMDREADY	;WAIT FOR COMMAND COMPLETION
	SHR	AL,01H		;TEST FOR ERROR CONDITION ON SEEK
	JNC	EXIT1
	STC
	MOV	DS:DSTAT,HBAD_SEEK	;BAD SEEK ERROR
EXIT1:				;RETURN
	RET
UPTASK	ENDP

PAGE
;---------------------------------------------------------------:
;HDIAG: Perform hard disk diagnostics.                          :
;       If error occurs, set BAD Controller error.              :
;                                                               :
;   ENTRY:NONE                                                  :
;    EXIT:CY=0 OK, CY=1 ERROR, DSTAT SET BAD CONTROLLER ERROR   :
;DESTROYS:AL                                                    :
;---------------------------------------------------------------:
;HDIAG	PROC	NEAR		;PERFORM HARD DISK DIAGNOSTICS
;	MOV	AL,0AH
;	OUT	SDH,AL		;SELECT HARD DISK DRIVE #0
;	MOV	AL,CTEST
;	OUT	COMND,AL	;OUTPUT TEST COMMAND
;	CALL	CMDREADY	;WAIT TILL FINISHED TEST
;	IN	AL,HARD_ERROR 	;GET ERROR REGISTER
;	CMP	AL,01H		;01H IS FDC MISSING, OK
				;ANY OTHER SET IS ERROR
;	JZ	EXIT2
;	MOV	DS:DSTAT,BAD_CNTLR	;ERROR IN CONTROLLER
;EXIT2:
;	RET
;HDIAG	ENDP

PAGE
;---------------------------------------------------------------:
;ERRCHK: Check for errors (very generalized).                   :
;       MODIFIES AH AND AL REGISTERS                            :
;                                                               :
;   ENTRY:NONE                                                  :
;    EXIT:CY=0 OK, CY=1 ERROR, DSTAT SET ACCORDINGLY            :
;DESTROYS:NONE                                                  :
;---------------------------------------------------------------:
ERRCHK	PROC	NEAR		;CHECK FOR ERRORS
	PUSH	AX
	XOR	AH,AH		;ZERO OUT AH, NO ERROR YET
	IN	AL,STATUS	;GET STATUS
	TEST	AL,04H		;TEST CORRECTED DATA BIT
	JZ	SKIP1
	MOV	AH,DATA_CORRECTED	;CORRECTED DATA HERE
SKIP1:
	SHR 	AL,01H		;TEST ERROR BIT
	JNC	EXIT3		;NO ERROR, EXIT
;
;	ERROR BIT IS SET:
;	INTERPRET ERROR FOR IBM BIOS LEVEL RETURN
;
	IN 	AL,HARD_ERROR	;GET ERROR ENTRY

	SHR 	AL,1
	MOV	AH,HBAD_ADDR_MARK	;DATA ADDRESS MARK NOT FOUND
	JC	EREXIT

	SHR	AL,1
	MOV	AH,INIT_FAIL  		;TR000 ERROR
	JC	EREXIT

 	SHR	AL,1
	MOV	AH,UNDEF_ERR       	;ABORTED COMMAND
	JC 	EREXIT

	SHR	AL,1
	SHR	AL,1
	MOV	AH,RECORD_NOT_FND	;ID NOT FOUND
	JC 	EREXIT

	SHR	AL,1
	MOV	AH,HBAD_ADDR_MARK      	;CRC ERROR ID FIELD          
	JC 	EREXIT

	SHR	AL,1
	MOV	AH,BAD_ECC             	;UNCORRECTABLE ERROR
	JC 	EREXIT

	SHR	AL,1
	MOV	AH,BAD_TRACK		;BAD BLOCK DETECT
	JC 	EREXIT

	MOV	AH,UNDEF_ERR		;UNDEFINED ERROR
EREXIT:
	MOV	AL,0		;NO SECTORS TRANSFERRED
	STC			;SET CARRY TO NOTIFY CALLING ROUTINE OF ERROR
COMMENT *
;BOB YOU SCREWED UP THE PATCH ! BS1/14/84
EXIT3:
	PUSH	DS		;SAVE DATA SEGMENT		:PATCH EC
	PUSH	AX		;SAVE AX			:PATCH EC
	MOV	AX,0040H	;CHANGE DATA SEGMENT TO 40H	:PATCH EC
	POP	AX		;RESTORE AX			:PATCH EC
	MOV	DS:DSTAT,AH	;PUT ERROR INTO DSTAT
	POP	DS		;RESTORE DATA SEGMENT		:PATCH EC

	POP	AX
	RET
ERRCHK	ENDP
*
; END OF ABOVE COMMENT
; THE SECTION BELOW REPLACES THE ABOVE CODE

EXIT3:
	PUSH	DS		;SAVE DATA SEGMENT		:PATCH EC
	PUSH	AX		;SAVE AX			:PATCH EC
	MOV	AX,0040H	;CHANGE DATA SEGMENT TO 40H	:PATCH EC
	MOV	DS,AX		;BS1/14/84                     
	POP	AX		;RESTORE AX			:PATCH EC
	MOV	DS:DSTAT,AH	;PUT ERROR INTO DSTAT
	POP	DS		;RESTORE DATA SEGMENT		:PATCH EC

	POP	AX
	RET
ERRCHK	ENDP

PAGE
;---------------------------------------------------------------:
;HDVERF: Verify sector transfer from Disk to Disk controller.   :
;                                                               :
;   ENTRY:                                                      :
;       CL = 2MSBits=CYLHI, 6LSBits=Start sector number         :
;       CH = Cylinder low                                       :
;       DL = Drive number                                       :
;       DH = Head number                                        :
;       AL = number of sectors to verify                        :
;                                                               :
;    EXIT: CY=0 OK, CY=1 ERROR                                  :
;                                                               :
;DESTROYS:  AL, CX, DX,                                         :
;---------------------------------------------------------------:
HDVERF	PROC	NEAR	
	CALL	UPTASK
	JC	EXIT4
	MOV	AL,1		;SET FOR ONE SECTOR READ AT A TIME
	OUT	SECNT,AL
VERF_LOOP:
	MOV	AL,DH	
	OUT	SECNO,AL	;SEND OUT SECTOR NUMBER TO HDC
	MOV	CX,512D		;SECTOR SIZE - POSSIBLY CHANGE FOR VARIANCE
	MOV	AL,CREAD
	OUT	COMND,AL	;READ SECTOR
	CALL	CMDREADY	;WAIT FOR CONTROLLER TO BE NOT BUSY
SEC1_LOOP:
	IN	AL,HDATA	;TURNQUIST: WRONG ADDRESS I/O
	LOOP	SEC1_LOOP
	CALL 	ERRCHK
	JC	EXIT4		;IF ERROR ....LET'S SPLIT

	INC	DH		;INCREMENT SECTOR NUMBER
;
;	SUPPORT FOR UP TO 80H SECTOR READS	11/26/84 BELOW RFS V V V V
;
	CMP	DH,10H
	JNZ	SKIPIT1
	MOV	DH,00H
	IN	AL,SDH
	INC	AL
	TEST	AL,04H
	JZ	SKIPIT1
	AND	AL,0FCH
	OUT	SDH,AL
	IN	AL,CYLLO
	INC	AL
	OUT	CYLLO,AL
	JNZ	SKIPIT1
	IN	AL,CYLHI
	INC	AL
	OUT	CYLHI,AL
SKIPIT1:
;
;	SUPPORT FOR UP TO 80H SECTOR READS	11/26/84 ABOVE RFS ^ ^ ^ ^
;
	DEC	DL		;DECREMENT SECTOR COUNT
	JNZ	VERF_LOOP
EXIT4:
	RET
HDVERF	ENDP

	
PAGE
;---------------------------------------------------------------:
;FMTTRK: Format specified track.                                :
;                                                               :
;  ENTRY:AH = TYPE OF FORMAT, =0 NORMAL, =80H BAD MARKING       :
;                                                               :
;  ALL PARAMETERS ARE FIXED. SECTORS/TRACK = 16D,               :
;  INTERLEAVE = 8:1 (ALA CP/M BIOS)                             :
;                                                               :
;---------------------------------------------------------------:
FMTTRK	PROC	NEAR
	PUSH    CX		;SAVE CX REGISTER
	MOV	AL,16D  	;16 SECTORS PER TRACK
	OUT	SECNT,AL	;# OF SECTORS PER TRACK
	MOV	AL,CFORM
	OUT	COMND,AL	;FORMAT COMMAND
;
;
	MOV	DL,0		;INTERLEAVE START SECTOR VALUE
	MOV	CX,16D		;LOAD CX WITH COUNT
LSET:
	MOV	AL,AH
	OUT  	HDATA,AL	;OUTPUT TYPE GOOD=0, BAD=80H
	MOV	AL,DL
	SHR	AL,1
	JNC	EVENN
	ADD	AL,8
EVENN:
	OUT	HDATA,AL	;LOAD INTERLEAVE 8:1
				; 0, 8, 1, 9, 2, A, 3, B, 4, C, 5, D,
	INC	DL		; 6, E, 7, F
	LOOP	LSET
;
	MOV	CX,0480D
GARBLOOP:			;SEND OUT REST OF 512 BYTE BUFFER FOR
	OUT	HDATA,AL	;CONTROLLERS BENEFIT.
	LOOP	GARBLOOP
;
;
	CALL	CMDREADY	;WAIT TILL CONTROLLER IS FINISHED
	POP	CX		;RESTORE CX REGISTER
	RET
FMTTRK ENDP

PAGE
;---------------------------------------------------------------:
;HFORMAT: Format track, or if cy is set, format to end of drive.:
;                                                               :
;       IF CARRY SET, FORMAT WHOLE TRACK                        :
;---------------------------------------------------------------:
HFORMAT	PROC	NEAR
	CALL	UPTASK
	JC	EXIT5		;ERROR IN LOADING TASK FILE

	PUSH	AX		;TURNQUIST VERSION: RFS 9/13/84
	MOV	AL,30D 		;TURNQUIST VERSION: RFS 9/13/84
	OUT	SECNO,AL	;TURNQUIST VERSION: RFS 9/13/84
	POP	AX		;TURNQUIST VERSION: RFS 9/13/84

	SAL	AH,1		;SHIFT MSBit to check for end of disk format
	JC	HFLOOP		;FORMAT TO END OF DISK
	CALL	FMTTRK		;FORMAT ONE TRACK ONLY
	CALL	ERRCHK
	JMP	EXIT5
HFLOOP:
	CMP	CX,0132H	;COMPARE W/MAXIMUM TRACK NUMBER
	JGE	EXIT5
	MOV	AL,CH
	OUT	CYLHI,AL
	MOV	AL,CL
	OUT	CYLLO,AL
	CALL	FMTTRK
	CALL	ERRCHK
	JC	EXIT5		;IF ERROR, ABORT REST OF OPERATION
	INC	CX
	JMP	HFLOOP
EXIT5:
	RET
HFORMAT	ENDP	

PAGE
;---------------------------------------------------------------:
;HDREAD: Read sector(s) into buffer.                            :
;ENTRY: AH = COMMAND (READ OR READLONG)                         :
;       CL = (2MSBits=CYLHI, 6LSBits=start sector number        :
;       CH = Cylinder low                                       :
;       DL = Drive number                                       :
;       DH = Head number                                        :
;	ES:BX=BUFFER ADDRESS					:
;---------------------------------------------------------------:
HDREAD	PROC	NEAR
	CALL	UPTASK		;LOAD THEM INTO THE HDC
	JC	EXIT6 
 	MOV	DI,BX		;BUFFER POINTER
	CLD
	MOV	BX,SECSIZE	;BX = SECTOR SIZE
	TEST	AH,02H		;CHECK FOR LONG READ (+4 BYTES TO SIZE)
	JZ	PREREAD
	ADD	BX,04H		;LONG READ INCLUDES 4 ECC BYTES
PREREAD:
	CALL	DMACHK		;CHECK FOR DMA BOUNDARY ERROR
	JC	EXIT6	
DOREAD:
	MOV	CX,BX		;AMOUNT OF DATA TO READ	(RLOOP COUNTER)_
	MOV	AL,DH		
	OUT	SECNO,AL	;NUMBER OF SECTOR TO READ
	MOV	AL,01H
	OUT	SECNT,AL	;ONE SECTOR AT A TIME ONLY
	MOV	AL,AH
	OUT	COMND,AL	;SEND OUT READ COMMAND
	CALL	CMDREADY
RLOOP:
	IN	AL,HDATA	;READ AND STORE BYTE
	STOSB
	LOOP	RLOOP
	CALL	ERRCHK		;CHECK FOR ERROR
	JC	EXIT6		;IF ERROR EXIT
	INC	DH		;INCREMENT SECTOR NUMBER READING	
;
;	SUPPORT FOR UP TO 80H SECTOR READS	11/26/84 BELOW RFS V V V V
;
	CMP	DH,10H
	JNZ	SKIPIT2
	MOV	DH,00H
	IN	AL,SDH
	INC	AL
	TEST	AL,04H
	JZ	SKIPIT2
	AND	AL,0FCH
	OUT	SDH,AL
	IN	AL,CYLLO
	INC	AL
	OUT	CYLLO,AL
	JNZ	SKIPIT2
	IN	AL,CYLHI
	INC	AL
	OUT	CYLHI,AL
SKIPIT2:
;
;	SUPPORT FOR UP TO 80H SECTOR READS	11/26/84 ABOVE RFS ^ ^ ^ ^
;
	DEC	DL		;DECREMENT COUNT OF SECTORS
	JNZ	DOREAD
EXIT6:
	RET
HDREAD	ENDP

PAGE
;---------------------------------------------------------------:
;HDWRITE: Write sector(s) from buffer to disk.                  :
;ENTRY: AH = COMMAND (WRITE OR WRITELONG)                       :
;       CL = (2MSBits=CYLHI, 6LSBits=start sector number        :
;       CH = Cylinder low                                       :
;       DL = Drive number                                       :
;       DH = Head number                                        :
;	ES:BX=BUFFER ADDRESS					:
;                                                               :
;---------------------------------------------------------------:
HDWRITE	PROC	NEAR
	PUSH	DS		;SAVE DATA SEGMENT
	CALL	UPTASK		;LOAD THEM INTO THE HDC
	JC	EXIT7
	PUSH	ES
	POP	DS
	MOV	SI,BX		;BUFFER POINTER
	CLD
	MOV	BX,SECSIZE	;BX = SECTOR SIZE
	TEST	AH,02H		;CHECK FOR LONG WRITE (+4 BYTES TO SIZE)
	JZ	PREWRITE
	ADD	BX,04H		;LONG WRITE INCLUDES 4 ECC BYTES
PREWRITE:
	CALL	DMACHK		;CHECK FOR DMA BOUNDARY ERROR
	JC	EXIT7
DOWRITE:
	MOV	CX,BX		;AMOUNT OF DATA TO WRITE(WLOOP COUNTER)
	MOV	AL,DH		
	OUT	SECNO,AL	;NUMBER OF SECTOR TO WRITE
	MOV	AL,01H
	OUT	SECNT,AL	;ONE SECTOR AT A TIME ONLY
	MOV	AL,AH
	OUT	COMND,AL	;SEND OUT WRITE COMMAND
WLOOP:
	LODSB
	OUT	HDATA,AL	;WRITE DATA TO HDC DATA PORT
	LOOP	WLOOP
	CALL	CMDREADY
	CALL	ERRCHK		;CHECK FOR ERROR
	JC	EXIT7		;IF ERROR EXIT
	INC	DH		;INCREMENT SECTOR NUMBER WRITING
;
;	SUPPORT FOR UP TO 80H SECTOR READS	11/26/84 BELOW RFS V V V V
;
	CMP	DH,10H
	JNZ	SKIPIT3
	MOV	DH,00H
	IN	AL,SDH
	INC	AL
	TEST	AL,04H
	JZ	SKIPIT3
	AND	AL,0FCH
	OUT	SDH,AL
	IN	AL,CYLLO
	INC	AL
	OUT	CYLLO,AL
	JNZ	SKIPIT3
	IN	AL,CYLHI
	INC	AL
	OUT	CYLHI,AL
SKIPIT3:
;
;	SUPPORT FOR UP TO 80H SECTOR READS	11/26/84 ABOVE RFS ^ ^ ^ ^
;
	DEC	DL		;DECREMENT COUNT OF SECTORS
	JNZ	DOWRITE
EXIT7:
	POP	DS		;RESTORE DATA SEGMENT
	RET
HDWRITE	ENDP


PAGE
;--------------------------------------------------------------:
;DMACHK: Check for 64K segment boundary overflow               :
;                                                              :
;If over then carry will be set to notify calling routine of   :
;error.                                                        :
;--------------------------------------------------------------:
DMACHK	PROC	NEAR
	PUSH	AX	;SAVE REGISTERS
	PUSH	DX
	SUB	DH,DH	;ZERO OUT DH REGISTER
	MOV	AX,CX	;BLOCKSIZE
	CLC
	MUL	DX	;DX = NUMBER OF BYTES TO TRANSFER
	JC	OVVER
	ADD	AX,BX	;ADD COUNT AND START ADDRESS FOR PAGE OVERFLOW TEST
	JNC 	ALRIGHT	;NO OVERFLOW, ALRIGHT
OVVER:
	MOV	DS:DSTAT,DMA_BOUNDARY
ALRIGHT:
	POP	DX
	POP	AX	;RESTORE REGISTERS
	RET
DMACHK	ENDP		;ROUTINE WILL RETURN CARRY SET ON ERROR

H1	LABEL	WORD		;FUNCTION TRANSFER TABLE
	DW	DISK_RESET	; 000H
	DW	RETURN_STATUS	; 001H
	DW	DISK_READ	; 002H
	DW	DISK_WRITE	; 003H
	DW	DISK_VERF	; 004H
	DW	FMT_TRK		; 005H
	DW	FMT_BAD		; 006H
	DW	FMT_DRV		; 007H
	DW	BAD_COMMAND	; 008H	(HANDLED INDEPENDENTLY OF TABLE)
	DW	INIT_DRV	; 009H
	DW	RD_LONG		; 00AH
	DW	WR_LONG		; 00BH
	DW	DISK_SEEK	; 00CH
	DW	DISK_RESET	; 00DH
	DW	RD_BUFF		; 00EH
	DW	WR_BUFF		; 00FH
	DW	TST_RDY		; 010H
	DW	HDISK_RECAL	; 011H
	DW	RAM_DIAG	; 012H
	DW	CHK_DRV		; 013H
	DW	CNTLR_DIAG	; 014H
H1L	EQU	$-H1

FD_TBL:
	DW	0306D		;MAXIMUM NUMBER OF CYLINDERS
	DB	04D		;MAXIMUM NUMBER OF HEADS
	DW	0000H		;STARTING REDUCED WRITE CURRENT CYL (N/A)
	DW	0000H		;STARTING WRITE PRECOMPENSATION CYL (N/A)
	DB	00H		;MAXIMUM ECC DATA BURST LENGTH (N/A)
	DB	00H		;CONTROL BYTE (N/A)
	DB	00H		;STANDARD TIME OUT VALUE (N/A)
	DB	00H		;TIME OUT VALUE FOR FORMAT DRIVE (N/A)
	DB	00H		;TIME OUT VALUE FOR CHECK DRIVE (N/A)
	DB	0,0,0,0		;RESERVED FOR FUTURE USE (N/A)

