		PAGE	60,132
;******************************************************************************
;
;				BOOT LOADER
;
;Description:
;
;	This code and data resides on the first physical sector of a floppy
;	disk, or on the first sector of the DOS partition of a hard disk.
;	The ROM BIOS loads this sector at location 0:7C00 from the floppy disk
;	in the boot drive, or if that fails, it transfers control to the master
;	boot record on the hard disk (if there).  If the partiton table in the 
;	master boot record contains a bootable DOS partition, then this loader
;	is loaded from the first sector of the partition and given control.
;
;	The purpose of the boot loader is to load the BIOS file (IBMBIO.COM)
;	and jump to the initialization routine in the BIOS.  To do this the
;	boot loader reads the first directory entry and makes sure it is the
;	BIOS file.  Then the boot loader reads in the BIOS file and jumps to
;	it.
;
;	If the BIOS file can not be loaded, an error message is displayed and
;	the program halts.
;
;	The data in this boot loader is set-up for double-sided, 9 sector per
;	track diskettes, and may be written as is to such a disk.  For hard
;	disks, certain data items must be modified according to partition size
;	and drive parameters.  The program which formats the hard disk must
;	alter and write this loader to the DOS partition after formatting.
;
;ENTRY:		NONE
;
;EXIT:		Jumps to BIOS init routine
;		AX	=	Drive code
;		CX	=	Drive code
;		BX	=	Start sector of first file (BIOS)
;		50:0	=	First directory sector
;
;******************************************************************************

;******************************************************************************
;
;EDIT HISTORY -
;
; To create a loader: (1) Edit this file, making changes, additions, etc.
;		      (2) assemble using ASM or MASM
;		      (3) link using LINK
;		      (4) change to bin file using EXE2BIN
;		      (5) save to 1st sector of disk using DEBUG as follows:
; 			   >DEBUG		(enter debug program)
;			   >n loader.bin	(name of file to load)
;			   >l			(load file)
;			   >w 7d00 0 0 1	(write data to drive 0 sector 0
;						 writing 1 sector)
;			   >q			(quit debug program)
;
;  8/25/83 - Original version created.
;
;  6/18/84 - changed disk parameter table for qx-88.
;
;  7/14/84 - cosmetic changes to loading/error message.
;
;  12/10/84 - Made compatible with IBM DOS boot loader data definition;
;             Generalized for floppy or hard disk bootability;
;             PAGE directives removed to save the forest
;
;******************************************************************************


SIGNADDR	=	7DFEH	; address of signature word
BOOTADDR	=	7C00H

;* NOTE -- The IBM BIOS assumes the
;  directory has been loaded at 
;  DIR_SEGMENT:DIR_ADDR -- do not 
;  change this address and expect this 
;  loader to work with the IBM BIOS 

DIR_SEGMENT	=	0050H		;Segment to read directory into
DIR_ADDR	=	0000H		;Start address for directory

LOAD_SEGMENT	=	0070H		;Segment to load file into
LOAD_ADDR	=	0000H		;Load file start address

ENTRY_LEN	=	20H		;Length of directory entry
LEN_FILENAME	=	0BH		;Length of filename in directory entry
FILE_SIZE	=	1CH		;Offset to file size in directory entry

DSK_INT		=	13H		;Disk interrupt
READ_CMD	=	02H		;Read command for disk interrupt routine

VIDINT		=	10H		;Video interrupt
DISPCOM		=	0EH		;Command to display a character

HARD_DISK	=	80H		;Hard disk drive code

CR		=	0DH		;Carriage return
LF		=	0AH		;Line feed

LOADED_CODE	SEGMENT AT LOAD_SEGMENT	;Set up segment to define address
		ORG	LOAD_ADDR	;  for BIOS load address
LOAD_START	LABEL	FAR

LOADED_CODE	ENDS


CODE		SEGMENT 
ASSUME		CS:CODE,SS:CODE,DS:CODE

		ORG	01EH*4
DISK_POINTER	LABEL	DWORD			;Pointer used by ROM
						;   points to table of disk information
		
		ORG	BOOTADDR

		JMP	START
		DB	'EPS 2.11'		;VERSION NUMBER 'DOS 2.11'

; Boot loader data area:
; (initialized for Double-sided, 9 sector floppies)
; (h) indicates must be changed for use on hard disk

SEC_LEN		DW	512		; Bytes per sector
CLUSTERS	DB	2		; (h) Sectors per cluster
RESERVED	DW	1		; (h) Number of reserved sectors
NUMFATS		DB	2		; Number of FATS
ROOT_ENTRIES	DW	112		; (h) Max # of root directory entries (7 sectors)
TOTAL_SEC	DW	40 * 9  * 2	; (h) Total number of sectors (including boot sector, directory, etc.) 
MEDIA		DB	0FDh		; (h) Media descriptor 
					;  (3 low bits: removable, not 8 sector/track, double sided)
FAT_LEN		DW	2		; (h) Number of sectors per FAT
SEC_PER_TRK	DW	9		; (h) Number of sectors per track
NUM_HEADS	DW	2		; (h) Number of heads
NUM_HIDDEN	DW	0		; (h) Number of hidden sectors (partitioned)

		;STUPID STUFF USED BY IBMBIO
DRIVE_TYPE	DB	0		; (h) Drive
		DB	0 		; Side
		DB	0AH		; Number of sectors to read

		;DISK TABLE USED BY ROM
DISK_BASE	LABEL	BYTE
							;Specify bytes for NEC
		DB	11001111B				;  D = 6 ms step rate, 
							;  F = 240 ms head unload time
		DB	2				;  1 ms head load time, DMA mode
		DB	37				;Motor wait-na qx-88
		DB	02h				;Bytes per sector:  00 = 128
							;		    01 = 256
							;		    02 = 512 etc
		DB	9				;EOT (Last sector on track)
		DB	02Ah				;Gap length
		DB	0FFh				;DTL
		DB	050h				;Gap length for format
		DB	0F6h				;Fill byte for format
		DB	25				;Head settle time (milliseconds)
		DB	4
			 
LOAD_FILENAME	DB	'IBMBIO   COM'			;Name of file to load
		
START:

;INITIALIZE
		CLI
		XOR	AX,AX
		MOV	DS,AX				;Initialize data segment
		MOV	SS,AX	
		MOV	SP,BOOTADDR			;Initialize stack pointer
		MOV	WORD PTR [DISK_POINTER],OFFSET DISK_BASE	;Set up disk pointer used by ROM to point to DISK_BASE
		MOV	WORD PTR [DISK_POINTER+2],CS
		STI
		INT	DSK_INT				;Reset disk to initialize new specify commands
							;  from DISK_BASE

;DISPLAY LOAD MESSAGE
		MOV 	SI,OFFSET LOADMESS
		CALL 	DISPMESS			;Display "Loading system ..." message
;
;CALCULATE START OF DIRECTORY AND START OF FIRST FILE
		MOV	AL,BYTE PTR [NUMFATS]		;AL = Number of FATs
		XOR     AH,AH                          
		MUL	FAT_LEN         		;AX = Number of FATs * number of sectors per FAT
		ADD	AX,RESERVED     		;Add reserved sectors at beginning of disk
		ADD	AX,NUM_HIDDEN			;Add offset to beginning of logical disk 
							;  (in case of partitioning)

		MOV	BX,AX				;BX = start sector of directory
		MOV	AX,ROOT_ENTRIES			;Maximum number of directory entries
		MOV	CX,ENTRY_LEN			;Bytes in directory
		MUL	CX
		XOR	DX,DX				;Clear DX for divide
		DIV	SEC_LEN				;AX = Number of sectors in directory
		ADD	AX,BX
		MOV	FILE_START,AX			;Save start sector of first file

;GET FIRST DIRECTORY ENTRY
		MOV	AX,BX				;AX = Start sector of directory
		MOV	CH,1				;Read 1 sector
		MOV	BX,DIR_SEGMENT
		MOV	ES,BX
		MOV	BX,DIR_ADDR			;ES:BX = DMA Address
		CALL	READ			 	
		JC	ERROR				;If read error

;VERIFY THAT FIRST FILE IS IBMBIO.COM
		MOV	SI,OFFSET LOAD_FILENAME		
		MOV	DI,LOAD_ADDR			;First filename in directory
		MOV	CX,LEN_FILENAME
		CMPSB					;Check if filenames match
		JNZ	ERROR				;If filenames don't match

;GET LENGTH OF IBMBIO.COM
		MOV	AX,ES:FILE_SIZE			;Get high word of file size (in bytes)
		MOV	DX,ES:FILE_SIZE+2		;Get low word of file size
		DIV	SEC_LEN				;AX = Number of sectors
		OR 	DX,DX				;IF NO REMAINDER 
		JZ	EVVEN				;...CONTINUE
		INC	AX				;...ELSE INC SECTORS
EVVEN:
 
;LOAD IBMBIO.COM	
		MOV	CH,AL				;CH = Number of sectors to read		
		MOV	AX,FILE_START			;AX = Start sector of file
		MOV	BX,LOAD_SEGMENT			
		MOV	ES,BX				
		MOV	BX,LOAD_ADDR			;ES:BX = DMA Address
	        CALL    READ			  	;Read the file
		JC	ERROR				;If read error

;SET UP NUMBER OF DRIVES AND START SECTOR OF FIRST FILE FOR IBMBIO.COM
		PUSH	CS
		POP	DS
		INT	11H				;Equipment check
		ROL	AL,1
		ROL	AL,1
		AND	AX,03H				;Get drive bits
		JNZ	DR

		INC	AX
DR:		INC	AX
		MOV	CX,AX				;CX = Number of logical floppies (2 for 1 or 2)
		CMP	DRIVE_TYPE,HARD_DISK		;Check if booted off of hard disk
							;Set AX = boot drive number
		JZ	D1				;If hard disk, boot drive = 2 = # of logical 
							;  floppies
		XOR	AX,AX				;If floppy, boot drive = 0
D1:
		PUSH	AX
		MOV	SI,OFFSET CLRMESS		;CLEAR OFF LOADING
		CALL	DISPMESS			;MESSAGE.
		POP	AX
		MOV	BX,FILE_START
		JMP	LOAD_START			;Jump to loaded program

ERROR:
		MOV	SI,OFFSET ERROR_MESS		
		CALL	DISPMESS
		XOR	AH,AH
	  	INT	16H		;GET CHAR FROM KEYBOARD
		INT	19H		;TRY TO BOOT AGAIN 

;----------------------------------------------------------
;READ -- Read sectors
;
;ENTRY:		AX	=	Logical sector number
;		CH	=	Number of sectors to read
;		ES:BX 	=	DMA Address
;EXIT:		C flag set if error
;----------------------------------------------------------

READ		PROC

	;CALCULATE INITIAL TRACK, SECTOR, AND HEAD
		PUSH	CX				; save # sectors
		XOR	DX,DX				;Clear for divide
		DIV	SEC_PER_TRK			;AX = # of Track * # of heads
							;DX = Sector number
		MOV	CL,DL				;CL = Sector number
		INC	CL				;CL = Actual sector number (count starting with 1)
			XOR	DX,DX	
		DIV	NUM_HEADS			;AX = Track number
							;DX = Head number

		MOV	CH,AL				;CH = Track number

; in case this is a hard disk, put bits 8,9 of track in bits 6,7 of CL
		ROR	AH,1
		ROR	AH,1
		AND	AH,0C0H
		AND	CL,03FH
		OR	CL,AH

		MOV	DH,DL				;DH = Head number

		POP	AX			; restore # sectors (AH)
NEXT_TRK:
	;CALCULATE NUMBER OF SECTORS TO READ ON THIS TRACK
		MOV	AL,AH				;AL = Number of sectors to read
		MOV	DL,BYTE PTR [SEC_PER_TRK]
		SUB	DL,CL				;Subtract start sector
		INC	DL				;DL = Maximum number of sectors to read on this track

		CMP	AL,DL				;Check number of sectors to read
		JC	READ_SEC			;If all sectors are on this track,
							;  read them all

		MOV	AL,DL				;If too many sectors, only read the ones on this track


READ_SEC:	
	;READ THE SECTORS ON THIS TRACK
		SUB 	AH,AL				;AH = Number of sectors left to read
		PUSH	AX				;Save AH = number of sectors left 
							; and AL = number read this time
		MOV	DL,BYTE PTR [DRIVE_TYPE]	;DL = Drive
		MOV	AH,READ_CMD			;Indicate read	
		INT	DSK_INT				;Read the sectors
		POP	AX				
		JC	READ_DONE			;Return error

	;CHECK IF ALL SECTORS READ
		OR	AH,AH				;Check if done
		JZ	READ_DONE			;Return no error

	;MOVE BUFFER POINTER TO NEXT SPOT IN BUFFER
		PUSH	DX
ADDSEC:		MOV	DX,SEC_LEN			;Add sector count
		ADD	BX,DX
		DEC	AL				;Decrement sector count
		JNZ	ADDSEC				;Add next sector
		POP	DX

	;SET UP NEXT TRACK, HEAD, AND SECTOR
		MOV	CL,1				;Sector 1
		INC	DH				;Next head
		CMP	DH,BYTE PTR [NUM_HEADS]		;Check if still more heads on this track
		JNZ	NEXT_TRK			;If so, go read sectors 

		XOR 	DH,DH				;Else, go to first head of next track
		INC	CH				;Next track
		JMP	NEXT_TRK			;Go read sectors from next track

READ_DONE:	RET

READ		ENDP

;----------------------------------------------------------
;DISPMESS -- Display message
;
;ENTRY:		SI	=	Index to message string
;				(string ending marked by 0)
;EXIT:		NONE
;----------------------------------------------------------

DISPMESS	PROC	
          	LODS	CS:BYTE PTR [SI]		;AL = next character in string
		OR	AL,AL				;Check for end of string
		JZ	DISP_DONE

		MOV	AH,DISPCOM			;Command to display the character in AL
		MOV	BH,0007
		INT	VIDINT				;Display character
		JMP	DISPMESS			;Get next character

DISP_DONE:	RET
DISPMESS       	ENDP
	
;***********************************************************
;
;			MESSAGES
;
;***********************************************************

LOADMESS	DB	'Loading System ... ',CR,0
CLRMESS		DB	'                   ',CR,0
ERROR_MESS	DB	'Non-System disk or disk error',CR,LF  
		DB	'Replace and strike any key when ready',CR,LF,0

;***********************************************************
;
;			VARIABLES
;
;***********************************************************
	
FILE_START	DW	0				;Holds start sector of load file

		ORG	SIGNADDR
		DB	055H,0AAH	

CODE		ENDS
		
		END
